@@ -153,6 +153,17 @@ static struct field_set_info super_fields[] = {
 	{ "backup_bgs", &set_sb.s_backup_bgs[0], NULL, 4, parse_uint,
 	  FLAG_ARRAY, 2 },
 	{ "checksum", &set_sb.s_checksum, NULL, 4, parse_uint },
+	{ "error_count", &set_sb.s_error_count, NULL, 4, parse_uint },
+	{ "first_error_time", &set_sb.s_first_error_time, NULL, 4, parse_time },
+	{ "first_error_ino", &set_sb.s_first_error_ino, NULL, 4, parse_uint },
+	{ "first_error_block", &set_sb.s_first_error_block, NULL, 8, parse_uint },
+	{ "first_error_func", &set_sb.s_first_error_func, NULL, 32, parse_string },
+	{ "first_error_line", &set_sb.s_first_error_ino, NULL, 4, parse_uint },
+	{ "last_error_time", &set_sb.s_last_error_time, NULL, 4, parse_time },
+	{ "last_error_ino", &set_sb.s_last_error_ino, NULL, 4, parse_uint },
+	{ "last_error_block", &set_sb.s_last_error_block, NULL, 8, parse_uint },
+	{ "last_error_func", &set_sb.s_last_error_func, NULL, 32, parse_string },
+	{ "last_error_line", &set_sb.s_last_error_ino, NULL, 4, parse_uint },
 	{ 0, 0, 0, 0 }
 };
 
@@ -310,8 +321,11 @@ static struct field_set_info *find_field(struct field_set_info *fields,
 	 */
 	if (idx) {
 		array_idx = strtol(idx, &tmp, 0);
-		if (*tmp)
+		if (*tmp) {
+			*(--idx) = '[';
+			*delim = ']';
 			idx = 0;
+		}
 	}
 
 	/*
@@ -543,8 +557,9 @@ static errcode_t parse_bmap(struct field_set_info *info,
 	}
 
 	retval = ext2fs_bmap2(current_fs, set_ino,
-			     (struct ext2_inode *) &set_inode,
-			     NULL, BMAP_SET, array_idx, NULL, &blk);
+			      (struct ext2_inode *) &set_inode,
+			      NULL, BMAP_ALLOC | BMAP_SET, array_idx, NULL,
+			      &blk);
 	if (retval) {
 		com_err("set_inode", retval, "while setting block map");
 	}
