@@ -7,5 +7,5 @@
  * file may be redistributed under the GNU Public License v2.
  */
 
-#define E2FSPROGS_VERSION "1.42.9"
-#define E2FSPROGS_DATE "28-Dec-2013"
+#define E2FSPROGS_VERSION "1.42.12"
+#define E2FSPROGS_DATE "29-Aug-2014"
